"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
var pretty_ms_1 = __importDefault(require("pretty-ms"));
var rollup = __importStar(require("rollup"));
var turbocolor_1 = __importDefault(require("turbocolor"));
var relativeId_1 = __importDefault(require("../../../src/utils/relativeId"));
var logging_1 = require("../logging");
var sourceMappingUrl_1 = __importDefault(require("../sourceMappingUrl"));
var timings_1 = require("./timings");
function build(inputOptions, outputOptions, warnings, silent) {
    if (silent === void 0) { silent = false; }
    var useStdout = outputOptions.length === 1 &&
        !outputOptions[0].file &&
        !outputOptions[0].dir &&
        inputOptions.input instanceof Array === false &&
        typeof inputOptions.input !== 'object';
    var start = Date.now();
    var files = useStdout ? ['stdout'] : outputOptions.map(function (t) { return relativeId_1.default(t.file || t.dir); });
    if (!silent) {
        var inputFiles = void 0;
        if (typeof inputOptions.input === 'string') {
            inputFiles = inputOptions.input;
        }
        else if (inputOptions.input instanceof Array) {
            inputFiles = inputOptions.input.join(', ');
        }
        else if (typeof inputOptions.input === 'object' && inputOptions.input !== null) {
            inputFiles = Object.keys(inputOptions.input)
                .map(function (name) { return inputOptions.input[name]; })
                .join(', ');
        }
        logging_1.stderr(turbocolor_1.default.cyan("\n" + turbocolor_1.default.bold(inputFiles) + " \u2192 " + turbocolor_1.default.bold(files.join(', ')) + "..."));
    }
    return rollup
        .rollup(inputOptions)
        .then(function (bundle) {
        if (useStdout) {
            var output_1 = outputOptions[0];
            if (output_1.sourcemap && output_1.sourcemap !== 'inline') {
                logging_1.handleError({
                    code: 'MISSING_OUTPUT_OPTION',
                    message: 'You must specify a --file (-o) option when creating a file with a sourcemap'
                });
            }
            return bundle.generate(output_1).then(function (_a) {
                var code = _a.code, map = _a.map;
                if (!code)
                    return;
                if (output_1.sourcemap === 'inline') {
                    code += "\n//# " + sourceMappingUrl_1.default + "=" + map.toUrl() + "\n";
                }
                process.stdout.write(code);
            });
        }
        return Promise.all(outputOptions.map(function (output) { return bundle.write(output); })).then(function () { return bundle; });
    })
        .then(function (bundle) {
        warnings.flush();
        if (!silent)
            logging_1.stderr(turbocolor_1.default.green("created " + turbocolor_1.default.bold(files.join(', ')) + " in " + turbocolor_1.default.bold(pretty_ms_1.default(Date.now() - start))));
        if (bundle && bundle.getTimings) {
            timings_1.printTimings(bundle.getTimings());
        }
    })
        .catch(function (err) {
        if (warnings.count > 0)
            warnings.flush();
        logging_1.handleError(err);
    });
}
exports.default = build;
