"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var path_1 = __importDefault(require("path"));
var rollup_1 = __importDefault(require("rollup"));
var turbocolor_1 = __importDefault(require("turbocolor"));
var relativeId_1 = __importDefault(require("../../../src/utils/relativeId"));
var logging_1 = require("../logging");
var batchWarnings_1 = __importDefault(require("./batchWarnings"));
function loadConfigFile(configFile, commandOptions) {
    if (commandOptions === void 0) { commandOptions = {}; }
    var silent = commandOptions.silent || false;
    var warnings = batchWarnings_1.default();
    return rollup_1.default
        .rollup({
        input: configFile,
        external: function (id) {
            return (id[0] !== '.' && !path_1.default.isAbsolute(id)) || id.slice(-5, id.length) === '.json';
        },
        onwarn: warnings.add
    })
        .then(function (bundle) {
        if (!silent && warnings.count > 0) {
            logging_1.stderr(turbocolor_1.default.bold("loaded " + relativeId_1.default(configFile) + " with warnings"));
            warnings.flush();
        }
        return bundle.generate({
            format: 'cjs'
        });
    })
        .then(function (_a) {
        var code = _a.code;
        // temporarily override require
        var defaultLoader = require.extensions['.js'];
        require.extensions['.js'] = function (module, filename) {
            if (filename === configFile) {
                module._compile(code, filename);
            }
            else {
                defaultLoader(module, filename);
            }
        };
        delete require.cache[configFile];
        return Promise.resolve(require(configFile))
            .then(function (configFileContent) {
            if (typeof configFileContent === 'function') {
                return configFileContent(commandOptions);
            }
            return configFileContent;
        })
            .then(function (configs) {
            if (Object.keys(configs).length === 0) {
                logging_1.handleError({
                    code: 'MISSING_CONFIG',
                    message: 'Config file must export an options object, or an array of options objects',
                    url: 'https://rollupjs.org/guide/en#configuration-files'
                });
            }
            require.extensions['.js'] = defaultLoader;
            return Array.isArray(configs) ? configs : [configs];
        });
    });
}
exports.default = loadConfigFile;
