"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
var date_time_1 = __importDefault(require("date-time"));
var fs_1 = __importDefault(require("fs"));
var pretty_ms_1 = __importDefault(require("pretty-ms"));
var rollup = __importStar(require("rollup"));
var signal_exit_1 = __importDefault(require("signal-exit"));
var turbocolor_1 = __importDefault(require("turbocolor"));
var mergeOptions_1 = __importDefault(require("../../../src/utils/mergeOptions"));
var relativeId_1 = __importDefault(require("../../../src/utils/relativeId"));
var logging_1 = require("../logging");
var alternateScreen_1 = __importDefault(require("./alternateScreen"));
var batchWarnings_1 = __importDefault(require("./batchWarnings"));
var loadConfigFile_1 = __importDefault(require("./loadConfigFile"));
var timings_1 = require("./timings");
function watch(configFile, configs, command, silent) {
    if (silent === void 0) { silent = false; }
    var isTTY = Boolean(process.stderr.isTTY);
    var warnings = batchWarnings_1.default();
    var initialConfigs = processConfigs(configs);
    var clearScreen = initialConfigs.every(function (config) { return config.watch.clearScreen !== false; });
    var screen = alternateScreen_1.default(isTTY && clearScreen);
    screen.open();
    var watcher;
    var configWatcher;
    var processConfigsErr;
    function processConfigs(configs) {
        return configs.map(function (options) {
            var merged = mergeOptions_1.default({
                config: options,
                command: command,
                defaultOnWarnHandler: warnings.add
            });
            var result = __assign(__assign({}, merged.inputOptions), { output: merged.outputOptions });
            if (!result.watch)
                result.watch = {};
            if (merged.deprecations.length) {
                result.watch._deprecations = merged.deprecations;
            }
            if (merged.optionError)
                merged.inputOptions.onwarn({
                    message: merged.optionError,
                    code: 'UNKNOWN_OPTION'
                });
            if (merged.inputOptions.watch &&
                merged.inputOptions.watch.clearScreen === false) {
                processConfigsErr = logging_1.stderr;
            }
            return result;
        });
    }
    function start(configs) {
        var screenWriter = processConfigsErr || screen.reset;
        watcher = rollup.watch(configs);
        watcher.on('event', function (event) {
            switch (event.code) {
                case 'FATAL':
                    screen.close();
                    logging_1.handleError(event.error, true);
                    process.exit(1);
                    break;
                case 'ERROR':
                    warnings.flush();
                    logging_1.handleError(event.error, true);
                    break;
                case 'START':
                    if (!silent) {
                        screenWriter(turbocolor_1.default.underline("rollup v" + rollup.VERSION));
                    }
                    break;
                case 'BUNDLE_START':
                    if (!silent) {
                        var input_1 = event.input;
                        if (typeof input_1 !== 'string') {
                            input_1 = Array.isArray(input_1)
                                ? input_1.join(', ')
                                : Object.keys(input_1)
                                    .map(function (key) { return input_1[key]; })
                                    .join(', ');
                        }
                        logging_1.stderr(turbocolor_1.default.cyan("bundles " + turbocolor_1.default.bold(input_1) + " \u2192 " + turbocolor_1.default.bold(event.output.map(relativeId_1.default).join(', ')) + "..."));
                    }
                    break;
                case 'BUNDLE_END':
                    warnings.flush();
                    if (!silent)
                        logging_1.stderr(turbocolor_1.default.green("created " + turbocolor_1.default.bold(event.output.map(relativeId_1.default).join(', ')) + " in " + turbocolor_1.default.bold(pretty_ms_1.default(event.duration))));
                    if (event.result && event.result.getTimings) {
                        timings_1.printTimings(event.result.getTimings());
                    }
                    break;
                case 'END':
                    if (!silent && isTTY) {
                        logging_1.stderr("\n[" + date_time_1.default() + "] waiting for changes...");
                    }
            }
        });
    }
    // catch ctrl+c, kill, and uncaught errors
    var removeOnExit = signal_exit_1.default(close);
    process.on('uncaughtException', close);
    // only listen to stdin if it is a pipe
    if (!process.stdin.isTTY) {
        process.stdin.on('end', close); // in case we ever support stdin!
    }
    function close(err) {
        removeOnExit();
        process.removeListener('uncaughtException', close);
        // removing a non-existent listener is a no-op
        process.stdin.removeListener('end', close);
        screen.close();
        if (watcher)
            watcher.close();
        if (configWatcher)
            configWatcher.close();
        if (err) {
            console.error(err);
            process.exit(1);
        }
    }
    try {
        start(initialConfigs);
    }
    catch (err) {
        close(err);
        return;
    }
    if (configFile && !configFile.startsWith('node:')) {
        var restarting_1 = false;
        var aborted_1 = false;
        var configFileData_1 = fs_1.default.readFileSync(configFile, 'utf-8');
        var restart_1 = function () {
            var newConfigFileData = fs_1.default.readFileSync(configFile, 'utf-8');
            if (newConfigFileData === configFileData_1)
                return;
            configFileData_1 = newConfigFileData;
            if (restarting_1) {
                aborted_1 = true;
                return;
            }
            restarting_1 = true;
            loadConfigFile_1.default(configFile, command)
                .then(function (_configs) {
                restarting_1 = false;
                if (aborted_1) {
                    aborted_1 = false;
                    restart_1();
                }
                else {
                    watcher.close();
                    start(initialConfigs);
                }
            })
                .catch(function (err) {
                logging_1.handleError(err, true);
            });
        };
        configWatcher = fs_1.default.watch(configFile, function (event) {
            if (event === 'change')
                restart_1();
        });
    }
}
exports.default = watch;
