"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var chunk_optimization_1 = require("../chunk-optimization");
var Graph_1 = __importDefault(require("../Graph"));
var addons_1 = require("../utils/addons");
var assetHooks_1 = require("../utils/assetHooks");
var assignChunkIds_1 = require("../utils/assignChunkIds");
var commondir_1 = __importDefault(require("../utils/commondir"));
var error_1 = require("../utils/error");
var fs_1 = require("../utils/fs");
var getExportMode_1 = __importDefault(require("../utils/getExportMode"));
var mergeOptions_1 = __importDefault(require("../utils/mergeOptions"));
var path_1 = require("../utils/path");
var sourceMappingURL_1 = require("../utils/sourceMappingURL");
var timers_1 = require("../utils/timers");
function addDeprecations(deprecations, warn) {
    var message = "The following options have been renamed \u2014 please update your config: " + deprecations
        .map(function (option) { return option.old + " -> " + option.new; })
        .join(', ');
    warn({
        code: 'DEPRECATED_OPTIONS',
        message: message,
        deprecations: deprecations
    });
}
function checkInputOptions(options) {
    if (options.transform || options.load || options.resolveId || options.resolveExternal) {
        throw new Error('The `transform`, `load`, `resolveId` and `resolveExternal` options are deprecated in favour of a unified plugin API. See https://rollupjs.org/guide/en#plugins');
    }
}
function checkOutputOptions(options) {
    if (options.format === 'es6') {
        error_1.error({
            message: 'The `es6` output format is deprecated – use `es` instead',
            url: "https://rollupjs.org/guide/en#output-format-f-format"
        });
    }
    if (!options.format) {
        error_1.error({
            message: "You must specify output.format, which can be one of 'amd', 'cjs', 'system', 'esm', 'iife' or 'umd'",
            url: "https://rollupjs.org/guide/en#output-format-f-format"
        });
    }
    if (options.moduleId) {
        if (options.amd)
            throw new Error('Cannot have both output.amd and output.moduleId');
    }
}
function getAbsoluteEntryModulePaths(chunks) {
    var absoluteEntryModulePaths = [];
    for (var _i = 0, chunks_1 = chunks; _i < chunks_1.length; _i++) {
        var chunk = chunks_1[_i];
        for (var _a = 0, _b = chunk.entryModules; _a < _b.length; _a++) {
            var entryModule = _b[_a];
            if (path_1.isAbsolute(entryModule.id)) {
                absoluteEntryModulePaths.push(entryModule.id);
            }
        }
    }
    return absoluteEntryModulePaths;
}
var throwAsyncGenerateError = {
    get: function () {
        throw new Error("bundle.generate(...) now returns a Promise instead of a { code, map } object");
    }
};
function applyOptionHook(inputOptions, plugin) {
    if (plugin.options)
        return plugin.options(inputOptions) || inputOptions;
    return inputOptions;
}
function getInputOptions(rawInputOptions) {
    if (!rawInputOptions) {
        throw new Error('You must supply an options object to rollup');
    }
    var _a = mergeOptions_1.default({
        config: rawInputOptions,
        deprecateConfig: { input: true }
    }), inputOptions = _a.inputOptions, deprecations = _a.deprecations, optionError = _a.optionError;
    if (optionError)
        inputOptions.onwarn({ message: optionError, code: 'UNKNOWN_OPTION' });
    if (deprecations.length)
        addDeprecations(deprecations, inputOptions.onwarn);
    checkInputOptions(inputOptions);
    var plugins = inputOptions.plugins;
    inputOptions.plugins = Array.isArray(plugins)
        ? plugins.filter(Boolean)
        : plugins
            ? [plugins]
            : [];
    inputOptions = inputOptions.plugins.reduce(applyOptionHook, inputOptions);
    if (!inputOptions.experimentalCodeSplitting) {
        inputOptions.inlineDynamicImports = true;
        if (inputOptions.manualChunks)
            error_1.error({
                code: 'INVALID_OPTION',
                message: '"manualChunks" option is only supported for experimentalCodeSplitting.'
            });
        if (inputOptions.experimentalOptimizeChunks)
            error_1.error({
                code: 'INVALID_OPTION',
                message: '"experimentalOptimizeChunks" option is only supported for experimentalCodeSplitting.'
            });
        if (inputOptions.input instanceof Array || typeof inputOptions.input === 'object')
            error_1.error({
                code: 'INVALID_OPTION',
                message: 'Multiple inputs are only supported for experimentalCodeSplitting.'
            });
    }
    if (inputOptions.inlineDynamicImports) {
        if (inputOptions.experimentalPreserveModules)
            error_1.error({
                code: 'INVALID_OPTION',
                message: "experimentalPreserveModules does not support the inlineDynamicImports option."
            });
        if (inputOptions.manualChunks)
            error_1.error({
                code: 'INVALID_OPTION',
                message: '"manualChunks" option is not supported for inlineDynamicImports.'
            });
        if (inputOptions.experimentalOptimizeChunks)
            error_1.error({
                code: 'INVALID_OPTION',
                message: '"experimentalOptimizeChunks" option is not supported for inlineDynamicImports.'
            });
        if (inputOptions.input instanceof Array || typeof inputOptions.input === 'object')
            error_1.error({
                code: 'INVALID_OPTION',
                message: 'Multiple inputs are not supported for inlineDynamicImports.'
            });
    }
    else if (inputOptions.experimentalPreserveModules) {
        if (inputOptions.manualChunks)
            error_1.error({
                code: 'INVALID_OPTION',
                message: 'experimentalPreserveModules does not support the manualChunks option.'
            });
        if (inputOptions.experimentalOptimizeChunks)
            error_1.error({
                code: 'INVALID_OPTION',
                message: 'experimentalPreserveModules does not support the experimentalOptimizeChunks option.'
            });
    }
    return inputOptions;
}
var curWatcher;
function setWatcher(watcher) {
    curWatcher = watcher;
}
exports.setWatcher = setWatcher;
function rollup(rawInputOptions) {
    try {
        var inputOptions_1 = getInputOptions(rawInputOptions);
        timers_1.initialiseTimers(inputOptions_1);
        var graph_1 = new Graph_1.default(inputOptions_1, curWatcher);
        curWatcher = undefined;
        // remove the cache option from the memory after graph creation (cache is not used anymore)
        var useCache_1 = rawInputOptions.cache !== false;
        delete inputOptions_1.cache;
        delete rawInputOptions.cache;
        timers_1.timeStart('BUILD', 1);
        return graph_1.pluginDriver
            .hookParallel('buildStart')
            .then(function () {
            return graph_1.build(inputOptions_1.input, inputOptions_1.manualChunks, inputOptions_1.inlineDynamicImports);
        })
            .then(function (chunks) {
            return graph_1.pluginDriver.hookParallel('buildEnd').then(function () {
                return chunks;
            });
        }, function (err) {
            return graph_1.pluginDriver.hookParallel('buildEnd', [err]).then(function () {
                throw err;
            });
        })
            .then(function (chunks) {
            timers_1.timeEnd('BUILD', 1);
            // TODO: deprecate legacy single chunk return
            var singleChunk;
            var singleInput = typeof inputOptions_1.input === 'string' ||
                (inputOptions_1.input instanceof Array && inputOptions_1.input.length === 1);
            if (!inputOptions_1.experimentalPreserveModules) {
                if (singleInput) {
                    for (var _i = 0, chunks_2 = chunks; _i < chunks_2.length; _i++) {
                        var chunk = chunks_2[_i];
                        if (singleChunk) {
                            singleChunk = undefined;
                            break;
                        }
                        singleChunk = chunk;
                    }
                }
            }
            // ensure we only do one optimization pass per build
            var optimized = false;
            function generate(rawOutputOptions, isWrite) {
                var outputOptions = normalizeOutputOptions(inputOptions_1, rawOutputOptions);
                if (inputOptions_1.experimentalCodeSplitting) {
                    if (typeof outputOptions.file === 'string' && typeof outputOptions.dir === 'string')
                        error_1.error({
                            code: 'INVALID_OPTION',
                            message: 'Build must set either output.file for a single-file build or output.dir when generating multiple chunks.'
                        });
                    if (chunks.length > 1) {
                        if (outputOptions.format === 'umd' || outputOptions.format === 'iife')
                            error_1.error({
                                code: 'INVALID_OPTION',
                                message: 'UMD and IIFE output formats are not supported with the experimentalCodeSplitting option.'
                            });
                        if (outputOptions.sourcemapFile)
                            error_1.error({
                                code: 'INVALID_OPTION',
                                message: '"sourcemapFile" is only supported for single-file builds.'
                            });
                    }
                    if (!singleChunk && typeof outputOptions.file === 'string')
                        error_1.error({
                            code: 'INVALID_OPTION',
                            message: singleInput
                                ? 'When building a bundle using dynamic imports, the output.dir option must be used, not output.file. Alternatively set inlineDynamicImports: true to output a single file.'
                                : 'When building multiple entry point inputs, the output.dir option must be used, not output.file.'
                        });
                }
                if (!outputOptions.file && inputOptions_1.experimentalCodeSplitting)
                    singleChunk = undefined;
                timers_1.timeStart('GENERATE', 1);
                var assetFileNames = outputOptions.assetFileNames || 'assets/[name]-[hash][extname]';
                var outputBundle = graph_1.finaliseAssets(assetFileNames);
                var inputBase = commondir_1.default(getAbsoluteEntryModulePaths(chunks));
                return graph_1.pluginDriver
                    .hookParallel('renderStart')
                    .then(function () { return addons_1.createAddons(graph_1, outputOptions); })
                    .then(function (addons) {
                    // pre-render all chunks
                    for (var _i = 0, chunks_3 = chunks; _i < chunks_3.length; _i++) {
                        var chunk = chunks_3[_i];
                        if (!inputOptions_1.experimentalPreserveModules)
                            chunk.generateInternalExports(outputOptions);
                        if (chunk.facadeModule && chunk.facadeModule.isEntryPoint)
                            chunk.exportMode = getExportMode_1.default(chunk, outputOptions);
                    }
                    for (var _a = 0, chunks_4 = chunks; _a < chunks_4.length; _a++) {
                        var chunk = chunks_4[_a];
                        chunk.preRender(outputOptions, inputBase);
                    }
                    if (!optimized && inputOptions_1.experimentalOptimizeChunks) {
                        chunk_optimization_1.optimizeChunks(chunks, outputOptions, inputOptions_1.chunkGroupingSize, inputBase);
                        optimized = true;
                    }
                    assignChunkIds_1.assignChunkIds(chunks, singleChunk, inputOptions_1, outputOptions, inputBase, addons);
                    var _loop_1 = function (i) {
                        var chunk = chunks[i];
                        var facadeModule = chunk.facadeModule;
                        outputBundle[chunk.id] = {
                            code: undefined,
                            dynamicImports: chunk.getDynamicImportIds(),
                            exports: chunk.getExportNames(),
                            facadeModuleId: facadeModule && facadeModule.id,
                            fileName: chunk.id,
                            imports: chunk.getImportIds(),
                            isDynamicEntry: facadeModule !== null && facadeModule.isDynamicEntryPoint,
                            isEntry: facadeModule !== null && facadeModule.isEntryPoint,
                            map: undefined,
                            modules: chunk.renderedModules,
                            get name() {
                                return chunk.getChunkName();
                            }
                        };
                    };
                    // assign to outputBundle
                    for (var i = 0; i < chunks.length; i++) {
                        _loop_1(i);
                    }
                    return Promise.all(chunks.map(function (chunk) {
                        var outputChunk = outputBundle[chunk.id];
                        return chunk.render(outputOptions, addons, outputChunk).then(function (rendered) {
                            outputChunk.code = rendered.code;
                            outputChunk.map = rendered.map;
                            return graph_1.pluginDriver.hookParallel('ongenerate', [
                                __assign({ bundle: outputChunk }, outputOptions),
                                outputChunk
                            ]);
                        });
                    })).then(function () { });
                })
                    .catch(function (error) {
                    return graph_1.pluginDriver.hookParallel('renderError', [error]).then(function () {
                        throw error;
                    });
                })
                    .then(function () {
                    // run generateBundle hook
                    // assets emitted during generateBundle are unique to that specific generate call
                    var assets = new Map(graph_1.assetsById);
                    var generateAssetPluginHooks = assetHooks_1.createAssetPluginHooks(assets, outputBundle, assetFileNames);
                    return graph_1.pluginDriver
                        .hookSeq('generateBundle', [outputOptions, outputBundle, isWrite], function (context) { return (__assign(__assign({}, context), generateAssetPluginHooks)); })
                        .then(function () {
                        // throw errors for assets not finalised with a source
                        assets.forEach(function (asset) {
                            if (asset.fileName === undefined)
                                assetHooks_1.finaliseAsset(asset, outputBundle, assetFileNames);
                        });
                    });
                })
                    .then(function () {
                    timers_1.timeEnd('GENERATE', 1);
                    return outputBundle;
                });
            }
            var cache = useCache_1 ? graph_1.getCache() : undefined;
            var result = {
                cache: cache,
                watchFiles: Object.keys(graph_1.watchFiles),
                generate: (function (rawOutputOptions) {
                    var promise = generate(rawOutputOptions, false).then(function (result) {
                        return inputOptions_1.experimentalCodeSplitting
                            ? { output: result }
                            : result[chunks[0].id];
                    });
                    Object.defineProperty(promise, 'code', throwAsyncGenerateError);
                    Object.defineProperty(promise, 'map', throwAsyncGenerateError);
                    return promise;
                }),
                write: (function (outputOptions) {
                    if (inputOptions_1.experimentalCodeSplitting &&
                        (!outputOptions || (!outputOptions.dir && !outputOptions.file))) {
                        error_1.error({
                            code: 'MISSING_OPTION',
                            message: 'You must specify output.file or output.dir for the build.'
                        });
                    }
                    else if (!inputOptions_1.experimentalCodeSplitting &&
                        (!outputOptions || !outputOptions.file)) {
                        error_1.error({
                            code: 'MISSING_OPTION',
                            message: 'You must specify output.file.'
                        });
                    }
                    return generate(outputOptions, true).then(function (outputBundle) {
                        return Promise.all(Object.keys(outputBundle).map(function (chunkId) {
                            return writeOutputFile(graph_1, result, chunkId, outputBundle[chunkId], outputOptions);
                        })).then(function () {
                            return inputOptions_1.experimentalCodeSplitting
                                ? { output: outputBundle }
                                : outputBundle[chunks[0].id];
                        });
                    });
                })
            };
            if (!inputOptions_1.experimentalCodeSplitting) {
                result.imports = singleChunk.getImportIds();
                result.exports = singleChunk.getExportNames();
                result.modules = (cache || graph_1.getCache()).modules;
            }
            if (inputOptions_1.perf === true)
                result.getTimings = timers_1.getTimings;
            return result;
        });
    }
    catch (err) {
        return Promise.reject(err);
    }
}
exports.default = rollup;
function isOutputChunk(file) {
    return typeof file.code === 'string';
}
function writeOutputFile(graph, build, outputFileName, outputFile, outputOptions) {
    var filename = path_1.resolve(outputOptions.dir || path_1.dirname(outputOptions.file), outputFileName);
    var writeSourceMapPromise;
    var source;
    if (isOutputChunk(outputFile)) {
        source = outputFile.code;
        if (outputOptions.sourcemap && outputFile.map) {
            var url = void 0;
            if (outputOptions.sourcemap === 'inline') {
                url = outputFile.map.toUrl();
            }
            else {
                url = path_1.basename(outputFileName) + ".map";
                writeSourceMapPromise = fs_1.writeFile(filename + ".map", outputFile.map.toString());
            }
            source += "//# " + sourceMappingURL_1.SOURCEMAPPING_URL + "=" + url + "\n";
        }
    }
    else {
        source = outputFile;
    }
    return fs_1.writeFile(filename, source)
        .then(function () { return writeSourceMapPromise; })
        .then(function () {
        return isOutputChunk(outputFile) &&
            graph.pluginDriver.hookSeq('onwrite', [
                __assign({ bundle: build }, outputOptions),
                outputFile
            ]);
    })
        .then(function () { });
}
function normalizeOutputOptions(inputOptions, rawOutputOptions) {
    if (!rawOutputOptions) {
        throw new Error('You must supply an options object');
    }
    // since deprecateOptions, adds the output properties
    // to `inputOptions` so adding that lastly
    var consolidatedOutputOptions = {
        output: __assign(__assign(__assign({}, rawOutputOptions), rawOutputOptions.output), inputOptions.output)
    };
    var mergedOptions = mergeOptions_1.default({
        // just for backward compatiblity to fallback on root
        // if the option isn't present in `output`
        config: consolidatedOutputOptions,
        deprecateConfig: { output: true }
    });
    if (mergedOptions.optionError)
        throw new Error(mergedOptions.optionError);
    // now outputOptions is an array, but rollup.rollup API doesn't support arrays
    var outputOptions = mergedOptions.outputOptions[0];
    var deprecations = mergedOptions.deprecations;
    if (deprecations.length)
        addDeprecations(deprecations, inputOptions.onwarn);
    checkOutputOptions(outputOptions);
    return outputOptions;
}
