"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var error_1 = require("./error");
var path_1 = require("./path");
var relativeId_1 = require("./relativeId");
function renderNamePattern(pattern, patternName, getReplacement) {
    if (!relativeId_1.isPlainName(pattern))
        error_1.error({
            code: 'INVALID_PATTERN',
            message: "Invalid output pattern \"" + pattern + "\" for " + patternName + ", cannot be an absolute or relative URL or path."
        });
    return pattern.replace(/\[(\w+)\]/g, function (_match, type) {
        var replacement = getReplacement(type);
        if (replacement === undefined)
            error_1.error({
                code: 'INVALID_PATTERN_REPLACEMENT',
                message: "\"" + type + "\" is not a valid substitution name in output option " + patternName + " pattern."
            });
        if (!relativeId_1.isPlainName(replacement))
            error_1.error({
                code: 'INVALID_PATTERN_REPLACEMENT',
                message: "Invalid replacement \"" + replacement + "\" for \"" + type + "\" in " + patternName + " pattern, must be a plain path name."
            });
        return replacement;
    });
}
exports.renderNamePattern = renderNamePattern;
function makeUnique(name, existingNames) {
    if (name in existingNames === false)
        return name;
    var ext = path_1.extname(name);
    name = name.substr(0, name.length - ext.length);
    var uniqueName, uniqueIndex = 1;
    while (existingNames[(uniqueName = name + ++uniqueIndex + ext)])
        ;
    return uniqueName;
}
exports.makeUnique = makeUnique;
